// agent.cc: agent implementation.
// 
// Copyright (C) 2004 Frank Langbein <frank@langbein.org>
// All rights reserved.
//
// This program is free software; you can redistribute it and/or modify it 
// under the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 of the License, or (at your option)
// any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 675 Mass Ave, Cambridge, MA 02139, USA.
// 
// $Id: main.cc,v 2.7 2003/09/01 12:30:26 scmfcl Exp $

#include "config.h"

RCSID (agent, "$Id$");

#include "xstring.hxx"
#include "list.hxx"

#include "agent.hxx"
#include "world.hxx"

char Agent::buffer[AGENT_BUFFER];
GLuint Agent::dl = 0;

Agent::Agent (unsigned long aid, string const &n, World *w,
	      SimpleList<Agent> *agts) : Object ("Agent_" + as_string (aid)) {
  id = aid;
  nm = n;

  scheme = false;
  unsigned int p = n.find ('.');
  unsigned int pp = 0;
  while (pp != string::npos) {
    pp = n.find ('.', p + 1);
    if (pp != string::npos)
      p = pp;
  }

  if (p != string::npos) {
    string ext = n.substr (p + 1, string::npos);
    if (ext == "scm") {
      scheme = true;
      func = n.substr (0, p);
      p = func.find ('/');
      pp = 0;
      while (pp != string::npos) {
	pp = func.find ('/', p + 1);
	if (pp != string::npos)
	  p = pp;
      }
      func = func.substr (p + 1, string::npos);
    }
  }

  string type;
  if (scheme) {
    gh_eval_file (nm.c_str ());
    type = "GAME"; // nothing else yet
  } else {
    int pfdout[2], pfdin[2];
    if (pipe (pfdout) == -1 || pipe (pfdin) == -1) {
      cerr << "agent " << nm << "(" << id 
	   << "): could not create pipelines" << endl;
      exit (1);
    }
    pid = fork ();
    switch (pid) {
    case -1:
      cerr << "agent " << nm << "(" << id << "): fork failed" << endl;
      exit (1);
    case 0:
      // Agent process
      if (dup2 (pfdout[0], 0) != 0) {
	cerr << "agent " << nm << "(" << id 
	     << "): write pipeline duplication failed" << endl;
	exit (1);
      }
      if (dup2 (pfdin[1], 1) != 1) {
	cerr << "agent " << nm << "(" << id
	     << "): read pipeline duplication failed" << endl;
	exit (1);
      }
      if (close (pfdout[0]) == -1 || close (pfdout[1]) == -1 ||
	  close (pfdin[0]) == -1 || close (pfdin[1]) == -1) {
	cerr << "agent " << nm << "(" << id 
	     << "): closing unused file descriptors failed"
	     << endl;
	exit (1);
      }
      execlp (nm.c_str (), nm.c_str (), NULL);
      cerr << "agent " << nm << "(" << id 
	   << "): could not execute agent" << endl;
      exit (1);
    }
    // World process
    if (close (pfdout[0]) == -1 || close(pfdin[1]) == -1) {
      cerr << "agent " << nm << "(" << id 
	   << "): closing unused file descriptors failed"
	   << endl;
      exit (1);
    }
    sndfp = fdopen (pfdout[1], "w");
    rcvfp = fdopen (pfdin[0], "r");
    string str;
    receive (str);
    int l = str.find (" ");
    type = str.substr (0, l);
  }

  if (w->is_game ()) {
    if (type != "GAME") {
      cerr << "agent " << nm << "(" << id 
	   << "): illegal initialisation string" << endl;
      exit (1);
    }
    if (!scheme)
      send ("SCHECKERS:" + as_string (w->sizeX()) + ":" + as_string (id));
  } else {
    if (type != "AGENT") {
      cerr << "agent " << nm << "(" << id 
	   << "): illegal initialisation string" << endl;
      exit (1);
    }
    bool oc = true;
    while (oc) {
      posx = static_cast<long> (static_cast<double> (w->sizeX ()) * RAND);
      posy = static_cast<long> (static_cast<double> (w->sizeY ()) * RAND);
      oc = ((*w) (posx, posy) -> length () > 0);
      if (!oc) {
	Agent *a = agts->first ();
	while (a) {
	  if (posx == a->positionX () &&
	      posy == a->positionY ()) {
	    oc = true;
	    break;
	  }
	  a = agts->next ();
	}
      }
    }
    o = static_cast<long> (4.0 * RAND);
    if (!scheme) 
      send ("WORLD:" + as_string (w->sizeX()) + ":" + as_string (w->sizeY ()));
  }

  carry = NULL;

  if (scheme) {
    cout << "Defined scheme agent " << nm << " (" << id << ")" 
	 << endl;
  } else {
    cout << "Initialised agent " << nm << " (" << id << "," << pid << ")" 
	 << endl;
  }

  broken = false;
}

Agent::~Agent () {
  if (carry)
    delete carry;
  if (!scheme) {
    send ("KILL");
    string str;
    receive (str);
    assert (str == "DYING");
    fclose (sndfp);
    fclose (rcvfp);
    kill (pid, SIGTERM);
    kill (pid, SIGKILL);
    waitpid (pid, NULL, WNOHANG);
  }
}

string Agent::name () const {
  return nm + "(" + as_string (id) + ")";
}

void Agent::send (string const &str) {
  assert (!scheme);
  if (fputs (str.c_str (), sndfp) == EOF) {
    cerr << "agent " << nm << "(" << id 
	 << "): write pipeline EOF (agent died?)"
	 << endl;
    exit(1);
  }
  fputc ('\n', sndfp);
  fflush (sndfp);
  cerr << "agent " << nm << "(" << id 
       << "): send " << str << endl;
  return;
}

void Agent::receive (string &str) {
  assert (!scheme);
  if (fgets (buffer, 127, rcvfp) == NULL) {
    cerr << "agent " << nm << "(" << id 
	 << "): reception failed" << endl;
    exit (1);
  }
  unsigned short l = 0;
  while (l < 127 && buffer[l] != '\0') {
    if (buffer[l] == '\n') {
      buffer[l] = '\0';
      break;
    }
    ++l;
  }
  str = static_cast<string> (buffer);
  cerr << "agent " << nm << "(" << id 
       << "): received " << str << endl;
  return;
}

GLfloat Agent::bdiffuse[] = {0.8, 0.8, 0.84, 1.0};
GLfloat Agent::bambient[] = {0.1, 0.1, 0.14, 1.0};
GLfloat Agent::bspecular[] = {0.8, 0.8, 0.84, 1.0};
GLfloat Agent::bshine = 127.0;

void Agent::init () {
  if (dl == 0) {
    dl = glGenLists (2);

    glNewList (dl, GL_COMPILE);

    glMaterialfv (GL_FRONT, GL_AMBIENT, bambient);
    glMaterialfv (GL_FRONT, GL_DIFFUSE, bdiffuse);
    glMaterialfv (GL_FRONT, GL_SPECULAR, bspecular);
    glMaterialfv (GL_FRONT, GL_SHININESS, &bshine);

    double c = M_PI / 180.0;
    for (long k = -30; k < 40; k += 5) {
      glBegin (GL_QUAD_STRIP);
      for (long l = 0; l <=360; l += 5) {
	glNormal3f (sin (l * c) * cos (k * c),
		    cos (l * c) * cos (k * c),
		    1.6 * sin (k * c));
	glVertex3f (sin (l * c) * cos (k * c) * 2.0,
		    cos (l * c) * cos (k * c) * 2.0,
		    1.6 * sin (k * c) * 2.0);
	long kk = k - 5;
	glNormal3f (sin (l * c) * cos (kk * c),
		    cos (l * c) * cos (kk * c),
		    1.6 * sin (kk * c));
	glVertex3f (sin (l * c) * cos (kk * c) * 2.0,
		    cos (l * c) * cos (kk * c) * 2.0,
		    1.6 * sin (kk * c) * 2.0);
      }
      glEnd ();
    }
    glNormal3f (0.0, 0.0, -1.0);
    glBegin (GL_TRIANGLE_FAN);
    double z = 1.6 * sin (-35 * c) * 2.0;
    glVertex3f (0.0, 0.0, z);
    for (long l = 0; l <= 360; l += 5)
      glVertex3f (sin (l * c) * cos (-35 * c) * 2.0,
		  cos (l * c) * cos (-35 * c) * 2.0,
		  z);
    
    glEnd ();
    glNormal3f (0.0, 0.0, 1.0);
    glBegin (GL_TRIANGLE_FAN);
    z = 1.6 * sin (35 * c) * 2.0;
    glVertex3f (0.0, 0.0, z);
    for (long l = 360; l >= 0; l -= 5)
      glVertex3f (sin (l * c) * cos (35 * c) * 2.0,
		  cos (l * c) * cos (35 * c) * 2.0,
		  z);
    
    glEnd ();

    glEndList ();

    glNewList (dl + 1, GL_COMPILE);

    glPushMatrix ();
   
    glTranslatef (0.0, 0.0, z);

    for (long k = 5; k < 30; k += 5) {
      glBegin (GL_QUAD_STRIP);
      for (long l = 0; l <= 360; l += 5) {
	glNormal3f (sin (l * c) * cos (k * c),
		    cos (l * c) * cos (k * c),
		    1.6 * sin (k * c));
	glVertex3f (sin (l * c) * cos (k * c) * .75 * 2.0,
		    cos (l * c) * cos (k * c) * .75 * 2.0,
		    1.6 * sin (k * c) * .75 * 2.0);
	long kk = k - 5;
	glNormal3f (sin (l * c) * cos (kk * c),
		    cos (l * c) * cos (kk * c),
		    1.6 * sin (kk * c));
	glVertex3f (sin (l * c) * cos (kk * c) * .75 * 2.0,
		    cos (l * c) * cos (kk * c) * .75 * 2.0,
		    1.6 * sin (kk * c) * .75 * 2.0);
      }
      glEnd ();
    }
    glNormal3f (0.0, 0.0, 1.0);
    glBegin (GL_TRIANGLE_FAN);
    z = 1.6 * sin (25 * c) * .75 * 2.0;
    glVertex3f (0.0, 0.0, z);
    for (long l = 360; l >= 0; l -= 5)
      glVertex3f (sin (l * c) * cos (25 * c) * .75 * 2.0,
		  cos (l * c) * cos (25 * c) * .75 * 2.0,
		  z);
    glEnd ();

    glPopMatrix ();
    
    for (unsigned long l = 0; l <=360; l += 120) {
      glPushMatrix ();
      glTranslatef (sin (l * c) * cos (-35 * c) * 2.0,
		    cos (l * c) * cos (-35 * c) * 2.0,
		    1.6 * sin (-35 * c) * 2.0);
      glutSolidSphere (.2 * 2.0, 10, 10);
      glPopMatrix ();
    }

    glEndList ();
  }

  double r = 0.0, g = 0.0, b = 0.0;
  switch (id % 6) {
  case 0:
    r = 1.0;
    g = 0.0;
    b = 0.0;
    break;
  case 1:
    r = 0.0;
    g = 1.0;
    b = 0.0;
    break;
  case 2:
    r = 0.0;
    g = 0.0;
    b = 1.0;
    break;
  case 3:
    r = 1.0;
    g = 1.0;
    b = 0.0;
    break;
  case 4:
    r = 1.0;
    g = 0.0;
    b = 1.0;
    break;
  case 5:
    r = 0.0;
    g = 1.0;
    b = 1.0;
    break;
  }
  double l = 1.0 / ((id / 6) + 1);
  
  ambient[0] = l * r * 0.1;
  ambient[1] = l * g * 0.1;
  ambient[2] = l * b * 0.1;
  ambient[3] = 1.0;
  diffuse[0] = l * r * 0.75;
  diffuse[1] = l * g * 0.75;
  diffuse[2] = l * b * 0.75;
  diffuse[3] = 1.0;
  specular[0] = l * r * 0.9;
  specular[1] = l * g * 0.9;
  specular[2] = l * b * 0.9;
  specular[3] = 1.0;

  return;
}

void Agent::render () const {
  glTranslatef (0.0, 0.0, (1.6 * sin (35 * M_PI / 180.0) + .2)* 2.0);
  
  glCallList (dl);

  glPushMatrix ();
  
  glTranslatef (sin (o * 90 * M_PI / 180.0) * cos (35 * M_PI / 180.0) *
		.7 * 2.0,
		cos (o * 90 * M_PI / 180.0) * cos (35 * M_PI / 180.0) *
		.7 * 2.0,
		2.5 * sin (35 * M_PI / 180.0) * 2.0);
  glutSolidSphere (.3 * 2.0, 10, 10);
  glPopMatrix ();

  glMaterialfv (GL_FRONT, GL_AMBIENT, ambient);
  glMaterialfv (GL_FRONT, GL_DIFFUSE, diffuse);
  glMaterialfv (GL_FRONT, GL_SPECULAR, specular);

  glCallList (dl + 1);
  return;
}

string scheme_command;

extern "C" {
  SCM cybot_make_move (SCM move) { // A hack...
    SCM from = gh_car (move);
    SCM to = gh_cadr (move);
    unsigned long f = gh_scm2ulong (from);
    unsigned long t = gh_scm2ulong (to);
    scheme_command = as_string (f) + "-" + as_string (t);
    return SCM_BOOL_T;
  }
}

Agent *Agent::turn (World *w) {
  Agent *zap = NULL;
  if (w->is_game ()) {

    unsigned long N = w->sizeX ();
    string cmd;

    if (scheme) {

      string board = "(list ";
      for (unsigned long r = 0; r < N; ++r) {
	for (unsigned long c = (r%2==0)?0:1; c < N; c+=2) {
	  Object *ob = (*w) (r, c)->first ();
	  if (!ob)
	    board += "'_ ";
	  else if (ob->name () == "Checkers_green") {
	    if (id == 0)
	      board += "'g ";
	    else
	      board += "'b ";
	  } else if (ob->name () == "Checkers_blue") {
	    if (id == 0)
	      board += "'b ";
	    else
	      board += "'g ";
	  } else {
	    board += "'_ ";
	  }
	}
      }
      board += ")";
      cmd = "(cybot_make_move (" + func + " " + as_string(N) + " " + board 
	+ "))";
      cerr << "agent " << nm << "(" << id << "): " << cmd << endl;
      gh_eval_str(cmd.c_str ());
      cmd = scheme_command;
      cerr << "agent " << nm << "(" << id << "): => " << cmd << endl;

    } else {
      string board = "";
      unsigned long N = w->sizeX ();
      for (unsigned long r = 0; r < N; ++r) {
	for (unsigned long c = (r%2==0)?0:1; c < N; c+=2) {
	  Object *ob = (*w) (r, c)->first ();
	  if (!ob)
	    board += '_';
	  else if (ob->name () == "Checkers_green") {
	    board += '0';
	  } else if (ob->name () == "Checkers_blue") {
	    board += '1';
	  } else {
	    board += '_';
	  }
	}
      }
      send (board);
      
      receive (cmd);
    }


    short win = 3;
    unsigned long from, to;
    if (sscanf (cmd.c_str (), "%d-%d", &from, &to) != 2) {
      cerr << "agent " << nm << "(" << id 
	   << "): illegal command" << endl;
      win = (id == 1) ? 0 : 1;
    } else {
      unsigned long fromr = (from * 2) / N;
      unsigned long fromc = (from * 2) % N;
      if ((N%2) == 0)
	fromc += (fromr % 2);
      Object *fob = (*w) (fromr, fromc)->first ();

      unsigned long tor = (to * 2) / N;
      unsigned long toc = (to * 2) % N;
      if ((N%2) == 0)
	toc += (tor % 2);
      Object *tob = (*w) (tor, toc)->first ();

      if (!fob || ((id == 0 && fob->name () != "Checkers_green") ||
		   (id == 1 && fob->name () != "Checkers_blue")) ||
	  tob) {
	cerr << "agent " << nm << "(" << id 
	     << "): illegal move" << endl;
	win = (id == 1) ? 0 : 1;
      } else {
	if (std::abs (static_cast<long> (fromr) - static_cast<long> (tor)) == 1
	    && std::abs (static_cast<long> (fromc) - static_cast<long> (toc)) == 1) {
	  (*w) (tor, toc) -> add_first ((*w) (fromr, fromc)->unlink_first ());
	  win = w->check_win (id);
	} else if (std::abs (static_cast<long> (fromr) - static_cast<long> (tor)) == 2 &&
		   std::abs (static_cast<long> (fromc) - static_cast<long> (toc)) == 2) {
	  long jr = (static_cast<long> (fromr) + static_cast<long> (tor)) / 2;
	  long jc = (static_cast<long> (fromc) + static_cast<long> (toc)) / 2;
	  Object *job = (*w) (jr, jc)->first ();
	  if (job && ((id == 0 && job->name () == "Checkers_blue") ||
		      (id == 1 && job->name () == "Checkers_green"))) {
	    (*w) (tor, toc) -> add_first ((*w) (fromr, fromc)->unlink_first ());
	    delete (*w) (jr, jc) -> unlink_first ();
	    win = w->check_win (id);
	  } else {
	    cerr << "agent " << nm << "(" << id 
		 << "): illegal move" << endl;
	    win = (id == 1) ? 0 : 1;
	  }
	} else {
	  cerr << "agent " << nm << "(" << id 
	       << "): illegal move" << endl;
	  win = (id == 1) ? 0 : 1;
	}
      }

    }

    if (win == 2) {
      cout << "Draw!" << endl;
      return NULL;
    }
    if (win != 3) {
      cout << "Agent " << win << " wins!" << endl;
      return NULL;
    }
    return this;

  } else {

    assert (!scheme);

    long oldposx = posx;
    long oldposy = posy;
    long sposx = posx;
    long sposy = posy;
    long dx = 0;
    long dy = 0;
    SimpleList<Object> *sol = NULL;
    switch (o) {
    case 0:
      sposy = posy + 1;
      dy = 1;
      if (static_cast<unsigned long> (posy + 1) < w->sizeY ())
	sol = (*w) (posx, posy + 1);
      break;
    case 1:
      sposx = posx + 1;
      dx = 1;
      if (static_cast<unsigned long> (posx + 1) < w->sizeX ())
	sol = (*w) (posx + 1, posy);
      break;
    case 2:
      sposy = posy - 1;
      dy = -1;
      if (posy > 0)
	sol = (*w) (posx, posy - 1);
    break;
    case 3:
      sposx = posx - 1;
      dx = -1;
      if (posx > 0)
	sol = (*w) (posx - 1, posy);
      break;
    }
    string sensor = "";
    if (sol) {
      Object *ob = sol->first ();
      while (ob) {
	if (sensor != "")
	  sensor += "/";
	sensor += ob->name ();
	ob = sol->next ();
      }
      if (sensor == "")
	sensor = "Empty";
    } else {
      sensor = "Border";
    }
    
    string atpos = "";
    SimpleList<Object> *ol = (*w) (posx, posy);
    Object *ob = ol->first ();
    while (ob) {
      if (ob != this) {
	if (atpos != "")
	  atpos += "/";
	atpos += ob->name ();
      }
      ob = ol->next ();
    }
    if (atpos == "")
      atpos = "Empty";
    
    send (atpos + ":" + sensor + ":" +
	  as_string (posx) + ":" + as_string (posy) + ":" + as_string (o));
    
    string cmd;
    receive (cmd);
    if (cmd == "COLLECT") {
      ob = ol->first ();
      Object *oo = ol->next ();
      while (oo && oo != this) {
	ob = oo;
	oo = ol->next ();
      }
      cout << ob->name () . substr (0, 7) <<endl;
      if (ob && ob->name () . substr (0, 7) == "Collect") {
	ol->del (ob);
	delete ob;
      }
    } else if (cmd == "TAKE") {
      if (carry) {
	cerr << "agent " << nm << "(" << id 
	     << "): already carries an object" << endl;
      } else {
	ob = sol->last ();
	if (ob && ob->name () . substr (0, 3) == "Box") {
	  sol->del (ob);
	  carry = ob;
	} else {
	  cerr << "agent " << nm << "(" << id 
	       << "): cannot take object" << endl;
	}
      }
    } else if (cmd == "DROP") {
      if (carry) {
	bool a = false;
	ob = sol->first ();
	while (ob) {
	  if (ob->name () . substr (0, 5) == "Agent") {
	    a = true;
	    break;
	  }
	  ob = sol->next ();
	}
	if (a) {
	  cerr << "agent " << nm << "(" << id 
	       << "): cannot drop object on agent" << endl;
	} else {
	  sol->add_last (carry);
	  carry = NULL;
	}
      } else {
	cerr << "agent " << nm << "(" << id 
	     << "): no object to drop" << endl;
      }
    } else if (cmd == "MOVE") {
      ob = sol->last ();
      if (ob && ob->name () . substr (0, 3) == "Box") {
	long ssposx = sposx + dx;
	long ssposy = sposy + dy;
	if (ssposx < 0 || static_cast<unsigned long> (ssposx) >= w->sizeX () ||
	    ssposy < 0 || static_cast<unsigned long> (ssposy) >= w->sizeY () ||
	    (*w) (ssposx, ssposy) -> length () != 0) {
	  cerr << "agent " << nm << "(" << id 
	       << "): cannot move object" << endl;
	} else {
	  sol->del (ob);
	  (*w) (ssposx, ssposy) -> add_last (ob);
	}
      } else {
	cerr << "agent " << nm << "(" << id 
	     << "): no object to move" << endl;
      }
    } else if (cmd == "ZAP") {      
      Agent *za = NULL;
      ob = sol->first ();
      while (ob) {
	if (ob->name () . substr (0, 5) == "Agent") {
	  za = dynamic_cast<Agent*> (ob);
	  assert (za);
	  break;
	}
	ob = sol->next ();
      }
      if (za) {
	sol->del (za);
	zap = za;
      } else {
	cerr << "agent " << nm << "(" << id 
	     << "): no agent to zap" << endl;
      }
    } else if (cmd == "X+") {
      ++posx;
    } else if (cmd == "X-") {
      --posx;
    } else if (cmd == "Y+") {
      ++posy;
    } else if (cmd == "Y-") {
      --posy;
    } else if (cmd == "R+") {
      ++o;
      if (o > 3)
	o = 0;
    } else if (cmd == "R-") {
      if (o == 0)
	o = 3;
      else 
	--o;
    } else if (cmd == "DYING") {
      zap = this;
    }
    
    if (static_cast<unsigned long> (posx) >= w->sizeX () || posx < 0 ||
	static_cast<unsigned long> (posy) >= w->sizeY () || posy < 0) {
      
      cerr << "agent " << nm << "(" << id 
	   << "): dropped off platform" << endl;
      send ("KILL");
      string cmd;
      receive (cmd);
      zap = this;
      (*w) (oldposx, oldposy) -> del (this);
      
    } else {
      
      bool crash = false;
      bool trap = false;
      string co = "";
      SimpleList<Object> *ol = (*w) (posx, posy);
      Object *ob = ol->first ();
      while (ob) {
	if (ob != this) {
	  if (ob->name () . substr (0, 3) == "Box" ||
	      ob->name () . substr (0, 5) == "Agent" ||
	      ob->name () . substr (0, 5) == "Block") {
	    crash = true;
	    if (co != "")
	      co += "/";
	    co += ob->name ();
	  } else if (ob->name () . substr (0, 4) == "Trap") {
	    trap = true;
	  }
	}
	ob = ol-> next ();
      }
      if (trap) {
	cerr << "agent " << nm << "(" << id 
	     << "): trapped" << endl;
	send ("KILL");
	string cmd;
	receive (cmd);
	zap = this;
	(*w) (oldposx, oldposy) -> del (this);
      } else if (crash) {
	cerr << "agent " << nm << "(" << id 
	     << "): crashed into " << co << endl;
	posx = oldposx;
	posy = oldposy;
      } else {
	(*w) (oldposx, oldposy) -> del (this);
	(*w) (posx, posy) -> add_last (this);
      }

    }
  }

  return zap;
}

/// Local Variables:          ***
/// compile-command: "make"   ***
/// End:                      ***
