; checkers.scm

; Convert list representing board to integer board representation
;     x0 * 3^0 + x1 * 3^1 + x2 * 3^2 + ... + xM & 3^M
; where xL = 0 (empty), 1 (green; me), 2 (blue; opponent)
(define to-internal
  (lambda (board pos)
    (if (null? board) 0
	(+ (to-internal (cdr board) (* pos 3))
	   (* (case (car board) 
		((b) 2)
		((g) 1)
		(else 0))
	      pos)))))

; board representation
(define convert-board
  (lambda (N board)
    (list (to-internal board 1) N '())))

; Get exponent in the integer board representation for position r,c
(define coordinates-to-exponent
  (lambda (r c B)
    (+ (if (even? (cadr B))
	   (* r (/ (cadr B) 2))
	   (if (even? r) (* (/ r 2) (cadr B)) (/ (+ (* r (cadr B)) 1) 2)))
       (/ (- c (if (even? r) 0 1)) 2))))

; Get contents of position in board state
(define get-pos
  (lambda (r c B)
    (modulo (quotient (car B) (expt 3 (coordinates-to-exponent r c B))) 3)))

; Set contents of board at given position and return new board state
(define set-pos
  (lambda (r c v B M)
    (list (+ (car B) (* (- v (get-pos r c B)) 
			(expt 3 (coordinates-to-exponent r c B))))
	  (cadr B) M)))

; Moves
(define up-left
  (lambda (r c B)
    (let ((rr (+ r 1))
	  (cc (+ c 1)))
      (if (and (< rr (cadr B)) (< cc (cadr B)) (= (get-pos rr cc B) 0))
	  (list (set-pos rr cc (get-pos r c B) (set-pos r c 0 B '())
			 (list (coordinates-to-exponent r c B)
			       (coordinates-to-exponent rr cc B))))
	  (list)))))

(define up-right
  (lambda (r c B)
    (let ((rr (+ r 1))
	  (cc (- c 1)))
      (if (and (< rr (cadr B)) (>= cc 0) (= (get-pos rr cc B) 0))
	  (list (set-pos rr cc (get-pos r c B) (set-pos r c 0 B '())
			 (list (coordinates-to-exponent r c B)
			       (coordinates-to-exponent rr cc B))))
	  (list)))))

(define down-left
  (lambda (r c B)
    (let ((rr (- r 1))
	  (cc (+ c 1)))
      (if (and (>= rr 0) (< cc (cadr B)) (= (get-pos rr cc B) 0))
	  (list (set-pos rr cc (get-pos r c B) (set-pos r c 0 B '())
			 (list (coordinates-to-exponent r c B)
			       (coordinates-to-exponent rr cc B))))
	  (list)))))

(define down-right
  (lambda (r c B)
    (let ((rr (- r 1))
	  (cc (- c 1)))
      (if (and (>= rr 0) (>= cc 0) (= (get-pos rr cc B) 0))
	  (list (set-pos rr cc (get-pos r c B) (set-pos r c 0 B '())
			 (list (coordinates-to-exponent r c B)
			       (coordinates-to-exponent rr cc B))))
	  (list)))))

(define jump-up-left
  (lambda (r c B)
    (let ((jrr (+ r 1))
	  (jcc (+ c 1))
	  (rr (+ r 2))
	  (cc (+ c 2)))
      (if (and (< rr (cadr B)) (< cc (cadr B)) (= (get-pos rr cc B) 0)
	       (= (get-pos jrr jcc B)
		  (if (= (get-pos r c B) 1) 2 1)))
	  (list (set-pos jrr jcc 0 
			 (set-pos rr cc (get-pos r c B) 
				  (set-pos r c 0 B '()) '())
			 (list (coordinates-to-exponent r c B)
			       (coordinates-to-exponent rr cc B))))
	  (list)))))
		
(define jump-up-right
  (lambda (r c B)
    (let ((jrr (+ r 1))
	  (jcc (- c 1))
	  (rr (+ r 2))
	  (cc (- c 2)))
      (if (and (< rr (cadr B)) (>= cc 0) (= (get-pos rr cc B) 0)
	       (= (get-pos jrr jcc B)
		  (if (= (get-pos r c B) 1) 2 1)))
	  (list (set-pos jrr jcc 0
			 (set-pos rr cc (get-pos r c B)
				  (set-pos r c 0 B '()) '())
			 (list (coordinates-to-exponent r c B)
			       (coordinates-to-exponent rr cc B))))
	  (list)))))

(define jump-down-left
  (lambda (r c B)
    (let ((jrr (- r 1))
	  (jcc (+ c 1))
	  (rr (- r 2))
	  (cc (+ c 2)))
      (if (and (>= rr 0) (< cc (cadr B)) (= (get-pos rr cc B) 0)
	       (= (get-pos jrr jcc B)
		  (if (= (get-pos r c B) 1) 2 1)))
	  (list (set-pos jrr jcc 0 
			 (set-pos rr cc (get-pos r c B)
				  (set-pos r c 0 B '()) '())
			 (list (coordinates-to-exponent r c B)
			       (coordinates-to-exponent rr cc B))))
	  (list)))))

(define jump-down-right
  (lambda (r c B)
    (let ((jrr (- r 1))
	  (jcc (- c 1))
	  (rr (- r 2))
	  (cc (- c 2)))
      (if (and (>= rr 0) (>= cc 0) (= (get-pos rr cc B) 0)
	       (= (get-pos jrr jcc B)
		  (if (= (get-pos r c B) 1) 2 1)))
	  (list (set-pos jrr jcc 0 
			 (set-pos r c (get-pos r c B)
				  (set-pos r c 0 B '()) '())
			 (list (coordinates-to-exponent r c B)
			       (coordinates-to-exponent rr cc B))))
	  (list)))))

; Compute successors for id player
(define successors
  (lambda (B id)
    (let ((succ '()))
      (let row_loop ((r 0))
	(let column_loop ((c (if (even? r) 0 1)))
	  (if (= (get-pos r c B) id)
	      (set! succ (append (up-left r c B)
				 (up-right r c B)
				 (down-left r c B)
				 (down-right r c B)
				 (jump-up-left r c B)
				 (jump-up-right r c B)
				 (jump-down-left r c B)
				 (jump-down-right r c B) succ)))
	  (if (< (+ c 2) (cadr B))
	      (column_loop (+ c 2))))
	(if (< (+ r 1) (cadr B))
	    (row_loop (+ r 1))))
      succ)))

; Evaluation function
(define eval
  (lambda (B)
    (let ((g 0)
	  (b 0))
      (let loop ((rb (car B)))
	(case (modulo rb 3)
	  ((1) (set! g (+ g 1)))
	  ((2) (set! b (+ b 1))))
	(if (> (quotient rb 3) 0) (loop (quotient rb 3))))
      (- g b))))

(define depth_limit (+ 2 (random 2)))

; Minimum successor
(define min-succ
  (lambda (B level list)
    (if (null? (cdr list))
	(max-node (car list) level)
	(min (max-node (car list) level) (min-succ B level (cdr list))))))

; Process min-node
(define min-node
  (lambda (B level)
    (let ((succs (successors B 2)))
      (cond ((null? succs) +10000)
	    ((> level depth_limit) (eval B))
	    (else (min-succ B (+ level 1) succs))))))

; Maximum successor
(define max-succ
  (lambda (B level list)
    (if (null? (cdr list))
	(min-node (car list) level)
	(max (min-node (car list) level) (max-succ B level (cdr list))))))

; Process max-node
(define max-node
  (lambda (B level)
    (let ((succs (successors B 1)))
      (cond ((null? succs) -10000)
	    ((> level depth_limit) (eval B))
	    (else (max-succ B (+ level 1) succs))))))

; Find move (max node reporting move rather than maximum alone)
(define move-succ
  (lambda (B list)
    (if (null? (cdr list))
	(cons (min-node (car list) 1) (car list))
	(let ((c (min-node (car list) 1))
	      (n (move-succ B (cdr list))))
	  (if (< c (car n)) n 
	      (if (> c (car n)) (cons c (car list))
		  (if (= (random 2) 0) n (cons c (car list)))))))))

(define move-node
  (lambda (B)
    (let ((succs (successors B 1)))
      (if (null? succs) (cons 1 (list 1 1 '(0 0)))
	  (move-succ B succs)))))

; Main function
(define checkers
  (lambda (N board)
    (caddr (cdr (move-node (convert-board N board))))))

; TODO: Seletive deepening
; TODO: alpha-beta
; TODO: general speed-up
