// main.cc: Main function for cybot.
//
// Copyright (C) 2004 Frank Langbein <frank@langbein.org>
// All rights reserved.
//
// This program is free software; you can redistribute it and/or modify it 
// under the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 of the License, or (at your option)
// any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 675 Mass Ave, Cambridge, MA 02139, USA.
//
// $Id: main.cc,v 2.7 2003/09/01 12:30:26 scmfcl Exp $

#include "config.h"

RCSID (main, "$Id$");

#include "list.hxx"

#include "version.hxx"
#include "engine.hxx"
#include "world.hxx"
#include "agent.hxx"

double EPSILON = DEFAULT_EPSILON;

// Not-a-number workaround test function.
//
// If we couldn't find any "not--a--number" (NaN) related code, we
// have to emulate it. This function checks if x is NaN by comparing
// it's bytes with the lilit_nan array, which represents the NaN code
// according to the IEEE standard.
#ifdef USE_SIL_ISNAN
extern "C" {
int isnan (double x)
{
  unsigned char *y = (unsigned char *) &x;
  if (y[0] == sil_nan[0] && y[1] == sil_nan[1] &&
      y[2] == sil_nan[2] && y[3] == sil_nan[3] &&
      y[4] == sil_nan[4] && y[5] == sil_nan[5] &&
      y[6] == sil_nan[6] && y[7] == sil_nan[7]) {
    return 1;
  }
  return 0;
}
}
#endif // USE_SIL_ISNAN

extern "C" {
  extern SCM cybot_make_move ();
}


void inner_main (int argc, char **argv)
{
  char *p = strrchr (argv[0], '/');
  if (!p)
    p = argv[0];
  else
    ++p;
  char *progname = new char[strlen (p) + 1];
  strcpy (progname, p);

  glutInit (&argc, argv);

  SimpleList<string> *agent_names = new SimpleList<string>;
  string world_type = "none";
  double sens_angle = DEFAULT_SENS_ANGLE;
  double sens_pos = DEFAULT_SENS_POSITION;
  bool error = false;
  int l = 1;
  while (l < argc) {
    if (strcmp (argv[l], "-v") == 0 ||
	strcmp (argv[l], "--version") == 0) {
      cout << "CyBot (" << progname << ") Version " << VERSION << " "
	   << " (" << TARGET_TYPE << ")\n";
      delete []progname;
      agent_names->purge ();
      delete agent_names;
      return;
    } else if (strcmp (argv[l], "-h") == 0 ||
	       strcmp (argv[l], "--help") == 0) {
      cerr << "Usage: " << progname << " {<options>} <agent1>...\n"
	   << " Options:\n"
	   << "   --help|-h                Print usage information\n"
	   << "   --version|-v             Print version information\n"
	   << "   --epsilon|-e EPS         Machine precision\n"
	   << "   --sensitivity|-s A,M     Angle and movement sensitivity\n"
	   << "   --world|-w WORLD         World type:\n"
	   << "                              vacuum:X:Y - vacuum cleaner world\n"
	   << "                              scheckers:N - simplified checkers\n"
	   << "\n"
	   << "Keyboard controls:\n"
	   << "  Forward/backward movement:     w,s\n"
	   << "  Strive left/right:             a,d\n"
	   << "  Move up/down:                  q,e\n"
	   << "  Single step agents:            x\n"
	   << "  Switch continuous agent steps: r\n"
	   << "  Left mouse button + movement:  look around\n"
	   << "  ESC:                           Quit\n";
      delete []progname;
      agent_names->purge ();
      delete agent_names;
      return;
    } else if (strcmp (argv[l], "-e") == 0 ||
	       strcmp (argv[l], "--epsilon") == 0) {
      ++l;
      if (l >= argc ||
	  sscanf (argv[l], "%lg", &EPSILON) != 1) {
	cerr << progname << ": illegal or missing precision argument" << endl;
	error = true;
      }
    } else if (strcmp (argv[l], "-s") == 0 ||
	       strcmp (argv[l], "--sensitivity") == 0) {
      ++l;
      if (l >= argc || 
	  sscanf (argv[l], "%lg,%lg", &sens_angle, &sens_pos) != 2) {
	cerr << progname << ": illegal or missing sensitivity arguments" << endl;
	error = true;
      }
    } else if (strcmp (argv[l], "-w") == 0 ||
	       strcmp (argv[l], "--world") == 0) {
      ++l;
      if (l >= argc) {
	cerr << progname << ": illegal world argument" << endl;
	error = true;
      } else if (world_type != "none") {
	cerr << progname << ": can only specify world type once" << endl;
	error = true;
      } else {
	world_type = static_cast<string> (argv[l]);
	int p = world_type.find (':');
	string type = world_type.substr (0, p);
	if (type != "vacuum" && type != "scheckers") {
	  cerr << progname << ": unkown world type " << world_type << endl;
	  error = true;
	}
      }
    } else {
      struct stat buf;
      if (stat (argv[l], &buf)) {
	cerr << progname << ": illegal argument or agent " << argv[l] << endl;
	error = true;
      } else {
	string *a = new string ();
	*a = static_cast<string> (argv[l]);
	agent_names->add_last (a);
      }
    }
    ++l;
  }

  if (agent_names->length () == 0) {
    cerr << progname << ": no agents specified" << endl;
    error = true;
  }
  if (world_type == "none") {
    cerr << progname << ": no world specified" << endl;
    error = true;
  }

  if (error) {
    cerr << progname << ": type `" << progname << " --help' for help\n";
  } else {
    srandom (time (NULL));
    gh_new_procedure ("cybot_make_move", cybot_make_move, 1, 0, 0);
    Engine *E = new Engine (new World (world_type), agent_names,
			    sens_pos, sens_angle);
    E->run ();
    delete E;
  }

  delete []progname;
  agent_names->purge ();
  delete agent_names;
  return;
}

int main (int argc, char **argv)
{
  gh_enter (argc, argv, inner_main);
  return (0); // never reached
}

/// Local Variables:        ***
/// compile-command: "make" ***
/// End:                    ***
