// world.cc: world implementation.
// 
// Copyright (C) 2004 Frank Langbein <frank@langbein.org>
// All rights reserved.
//
// This program is free software; you can redistribute it and/or modify it 
// under the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 of the License, or (at your option)
// any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 675 Mass Ave, Cambridge, MA 02139, USA.
// 
// $Id: main.cc,v 2.7 2003/09/01 12:30:26 scmfcl Exp $

#include "config.h"

RCSID (world, "$Id$");

#include "xstring.hxx"

#include "list.hxx"
#include "world.hxx"

GLfloat World::light_diffuse[]   = {1.0, 1.0, 1.0, 1.0};
GLfloat World::light_ambient[]   = {0.2, 0.2, 0.2, 1.0};
GLfloat World::light_specular[]  = {1.0, 1.0, 1.0, 1.0};
GLfloat World::light_attenuation = 1.2;

GLfloat World::floor_diffuse1[]  = {1.0, 1.0, 1.0, 1.0};
GLfloat World::floor_ambient1[]  = {0.2, 0.2, 0.2, 1.0};
GLfloat World::floor_specular1[] = {1.0, 1.0, 1.0, 1.0};
GLfloat World::floor_shine1      = 1.0;
GLfloat World::floor_diffuse2[]  = {0.0, 0.0, 0.0, 1.0};
GLfloat World::floor_ambient2[]  = {0.2, 0.2, 0.2, 1.0};
GLfloat World::floor_specular2[] = {0.6, 0.6, 0.6, 1.0};
GLfloat World::floor_shine2      = 100;

World::World (string world_type) {
  count = 0UL;
  unsigned int p = world_type.find (':');
  string basetype = world_type.substr (0, p);
  if (p != string::npos) {
    unsigned int pp = world_type.find (':', p + 1);
    if (pp != string::npos) {
      if (sscanf (world_type.substr (p + 1, string::npos) . c_str (),
		  "%d:%d", &sX, &sY) != 2) {
	sX = sY = 8;
      }
    } else {
      if (sscanf (world_type.substr (p + 1, string::npos) . c_str (),
		  "%d", &sX) != 1) {
	sX = sY = 8;
      } else {
	sY = sX;
      }
    }
  } else {
    sX = sY = 8;
  }
  grid = new SimpleList<Object>*[sX * sY];

  for (unsigned long l = 0; l < sX * sY; grid[l++] = new SimpleList<Object>);
  if (basetype == "vacuum") {
    vacuum_world ();
    g = false;
  } else if (basetype == "scheckers") {
    scheckers_world (); 
    g = true;
  } else {
    assert (false);
  }
  light_pos[0] = 0.0;
  light_pos[1] = 0.0;
  light_pos[2] = 15.0;
  light_pos[3] = 1.0;
  dlW = 0;
}

World::~World () {
  for (unsigned long l = 0; l < sX * sY; ++l) {
    grid[l]->purge ();
    delete grid[l];
  }
  delete []grid;
  if (dlW)
    glDeleteLists (dlW, 1);
}

void World::vacuum_world ()
{
  for (unsigned short x = 0; x < sX; ++x) {
    for (unsigned short y = 0; y < sY; ++y) {
      if (RAND < 0.3333) {
	(*this) (x, y) -> add_last (new Collect (BROWN));
      } else if (RAND < 0.25) {
	(*this) (x, y) -> add_last (new Box (CUBE, BLUE));
      }
    }
  }
  return;
}

void World::scheckers_world ()
{
  assert (sX > 6);
  for (unsigned short r = 0; r < 3; ++r) {
    for (unsigned short c = (r%2==0)?0:1; c < sX; c+=2) {
      (*this) (r, c) -> add_last (new Checkers (GREEN));
      (*this) (sX - 1 - r, sX - 1 - c) -> add_last (new Checkers (BLUE));
    }
  }
  return;
}

void World::init () {

  glEnable (GL_LIGHT1);
  glLightfv (GL_LIGHT1, GL_AMBIENT, light_ambient);
  glLightfv (GL_LIGHT1, GL_DIFFUSE, light_diffuse);
  glLightfv (GL_LIGHT1, GL_SPECULAR, light_specular);
  glLightfv (GL_LIGHT1, GL_CONSTANT_ATTENUATION, &light_attenuation);
  light_pos[0] = 2.0 * sX;
  light_pos[1] = 2.0 * sY;

  if (dlW)
    glDeleteLists (dlW, 1);
  dlW = glGenLists (1);

  glNewList (dlW, GL_COMPILE);

  glLightfv (GL_LIGHT1, GL_POSITION, light_pos);

  glNormal3f (0.0, 0.0, 1.0); 
  for (unsigned long x = 0; x < sX; ++x) {
    for (unsigned long y = 0; y < sY; ++y) {
      if ((x + y) % 2 == 0) {
	glMaterialfv (GL_FRONT, GL_AMBIENT, floor_ambient1);
	glMaterialfv (GL_FRONT, GL_DIFFUSE, floor_diffuse1);
	glMaterialfv (GL_FRONT, GL_SPECULAR, floor_specular1);
	glMaterialfv (GL_FRONT, GL_SHININESS, &floor_shine1);
      } else {
	glMaterialfv (GL_FRONT, GL_AMBIENT, floor_ambient2);
	glMaterialfv (GL_FRONT, GL_DIFFUSE, floor_diffuse2);
	glMaterialfv (GL_FRONT, GL_SPECULAR, floor_specular2);
	glMaterialfv (GL_FRONT, GL_SHININESS, &floor_shine2);
      }
      GLfloat cx = x * 4.0;
      GLfloat cy = y * 4.0;
      glBegin (GL_QUAD_STRIP);
      glVertex3f (cx - 2.0, cy - 2.0, 0.0);
      glVertex3f (cx - 2.0, cy + 2.0, 0.0);
      glVertex3f (cx + 2.0, cy - 2.0, 0.0);
      glVertex3f (cx + 2.0, cy + 2.0, 0.0);
      glEnd ();
    }
  }

  glEndList ();

  for (unsigned long x = 0; x < sX; ++x) {
    for (unsigned long y = 0; y < sY; ++y) {
      SimpleList<Object> *ol = (*this) (x, y);
      Object *o = ol->first ();
      while (o) {
	o->init ();
	o = ol->next ();
      }
    }
  }


  return;
}

void World::render () {
  glCallList (dlW);

  // Draw grid contents
  for (unsigned long x = 0; x < sX; ++x) {
    for (unsigned long y = 0; y < sY; ++y) {
      SimpleList<Object> *ol = (*this) (x, y);
      GLfloat h = 0.0;
      Object *o = ol->first ();
      while (o) {
	glPushMatrix ();
	glTranslatef (x * 4.0, y * 4.0, h);
	o->render ();
	glPopMatrix ();
	h += o->height ();
	o = ol->next ();
      }
    }
  }

  return;
}

int World::check_win (unsigned long id)
{
  bool can_move_0 = false;
  bool can_move_1 = false;
  unsigned long checkers_0 = 0UL;
  unsigned long checkers_1 = 0UL;

  for (unsigned short r = 0; r < sX; ++r) {
    for (unsigned short c = (r%2==0)?0:1; c < sX; c+=2) {
      Object *ob = (*this) (r, c) -> first ();
      if (ob) {
	if (ob->name () == "Checkers_green") {
	  ++checkers_0;
	  if (!can_move_0) {
	    if (r < sX - 1) {
	      if (c < sX - 1)
		can_move_0 |= ((*this) (r + 1, c + 1) -> length () == 0);
	      if (c > 0)
		can_move_0 |= ((*this) (r + 1, c - 1) -> length () == 0);
	    }
	    if (r > 0) {
	      if (c < sX - 1)
		can_move_0 |= ((*this) (r - 1, c + 1) -> length () == 0);
	      if (c > 0)
		can_move_0 |= ((*this) (r - 1, c - 1) -> length () == 0);
	    }
	    if (r < sX - 2) {
	      if (c < sX - 2) {
		Object *ob = (*this) (r + 1, c + 1) -> first (); 
		can_move_0 |= (ob && ob->name () == "Checkers_blue" &&
			       (*this) (r + 2, c + 2) -> length () == 0);
	      }
	      if (c > 1) {
		Object *ob = (*this) (r + 1, c - 1) -> first (); 
		can_move_0 |= (ob && ob->name () == "Checkers_blue" &&
			       (*this) (r + 2, c - 2) -> length () == 0);
	      }
	    }
	    if (r > 1) {
	      if (c < sX - 2) {
		Object *ob = (*this) (r - 1, c + 1) -> first (); 
		can_move_0 |= (ob && ob->name () == "Checkers_blue" &&
			       (*this) (r - 2, c + 2) -> length () == 0);
	      }
	      if (c > 1) {
		Object *ob = (*this) (r - 1, c - 1) -> first (); 
		can_move_0 |= (ob && ob->name () == "Checkers_blue" &&
			       (*this) (r - 2, c - 2) -> length () == 0);
	      }
	    }
	  }
	} else if (ob->name () == "Checkers_blue") {
	  ++checkers_1;
	  if (!can_move_1) {
	    if (r < sX - 1) {
	      if (c < sX - 1)
		can_move_1 |= ((*this) (r + 1, c + 1) -> length () == 0);
	      if (c > 0)
		can_move_1 |= ((*this) (r + 1, c - 1) -> length () == 0);
	    }
	    if (r > 0) {
	      if (c < sX - 1)
		can_move_1 |= ((*this) (r - 1, c + 1) -> length () == 0);
	      if (c > 0)
		can_move_1 |= ((*this) (r - 1, c - 1) -> length () == 0);
	    }
	    if (r < sX - 2) {
	      if (c < sX - 2) {
		Object *ob = (*this) (r + 1, c + 1) -> first (); 
		can_move_1 |= (ob && ob->name () == "Checkers_green" &&
			       (*this) (r + 2, c + 2) -> length () == 0);
	      }
	      if (c > 1) {
		Object *ob = (*this) (r + 1, c - 1) -> first (); 
		can_move_1 |= (ob && ob->name () == "Checkers_green" &&
			       (*this) (r + 2, c - 2) -> length () == 0);
	      }
	    }
	    if (r > 1) {
	      if (c < sX - 2) {
		Object *ob = (*this) (r - 1, c + 1) -> first (); 
		can_move_1 |= (ob && ob->name () == "Checkers_green" &&
			       (*this) (r - 2, c + 2) -> length () == 0);
	      }
	      if (c > 1) {
		Object *ob = (*this) (r - 1, c - 1) -> first (); 
		can_move_1 |= (ob && ob->name () == "Checkers_green" &&
			       (*this) (r - 2, c - 2) -> length () == 0);
	      }
	    }
	  }
	}
      }
    }
  }

  if (id == 0) {
    if (!can_move_1)
      return 0;
  } else if (id == 1) {
    if (!can_move_0)
      return 1;
  }

  if (id == 1) {
    ++count;
    if (count > sX * sX * sX) {
      cout << "Maximum number of plys reached\n";
      if (checkers_0 > checkers_1) {
	return 0;
      }
      if (checkers_1 > checkers_0) {
	return 1;
      }
      return 2;
    }
  }

  return 3;
}

GLfloat Object::diffuse[] = {0.8, 0.0, 0.0, 1.0,
			     0.0, 0.8, 0.0, 1.0,
			     0.0, 0.0, 0.8, 1.0,
			     0.6, 0.6, 0.0, 1.0,
			     0.6, 0.0, 0.6, 1.0,
			     0.0, 0.6, 0.6, 1.0,
			     0.5, 0.25, 0.0, 1.0,
                             0.8, 0.8, 0.8, 1.0,
                             0.0, 0.0, 0.0, 1.0 };
GLfloat Object::ambient[] = {0.2, 0.0, 0.0, 1.0,
			     0.0, 0.2, 0.0, 1.0,
			     0.0, 0.0, 0.2, 1.0,
			     0.2, 0.2, 0.0, 1.0,
			     0.2, 0.0, 0.2, 1.0,
			     0.0, 0.2, 0.2, 1.0,
			     0.1, 0.05, 0.0, 1.0,
                             0.3, 0.3, 0.3, 1.0,
                             0.0, 0.0, 0.0, 1.0 };
GLfloat Object::specular[] = {1.0, 0.0, 0.0, 1.0,
			      0.0, 1.0, 0.0, 1.0,
			      0.0, 0.0, 1.0, 1.0,
			      1.0, 1.0, 0.0, 1.0,
			      1.0, 0.0, 1.0, 1.0,
			      0.0, 1.0, 1.0, 1.0,
			      0.3, 0.2, 0.0, 1.0,
                              0.6, 0.6, 0.6, 1.0,
                              0.6, 0.6, 0.6, 1.0 };

GLfloat Object::shine = 50.0;


string Object::color[] = { "red", "green", "blue", "yellow", "purple",
			   "cyan", "brown", "white", "black" };
string Object::shape[] = { "Cube", "Cone", "Sphere", "Torus", "Teapot",
			   "Disc", "Cyl" };

GLuint Object::dl = 0;

Object::Object (string const &str) {
  nm = str;
}

void Object::setup () {
  if (!dl) {
    dl = glGenLists (7);
    for (unsigned short sh = 0; sh < 7; ++sh) {
      glNewList (dl + sh, GL_COMPILE);
      switch (sh) {
      case 0:
	glTranslatef (0, 0, 2.0);
	glutSolidCube (4.0);
	break;
      case 1:
	glTranslatef (0, 0, 2.0);
	glutSolidCone (4.0, 4.0, 20, 20);
	break;
      case 2:
	glTranslatef (0, 0, 2.0);
	glutSolidSphere (4.0, 20, 20);
	break;
      case 3:
	glTranslatef (0, 0, 2.0);
	glutSolidTorus (4.0, 2.0, 20, 20);
	break;
      case 4:
	glTranslatef (0, 0, 2.0);
	glutSolidTeapot (4.0);
	break;
      case 5:
	glNormal3f (0.0, 0.0, 1.0);
	glBegin (GL_POLYGON);
	glVertex3f (0.0, 1.0, 0.05);
	for (unsigned short l = 0; l < 20; ++l)
	  glVertex3f (sin (2.0 * M_PI - l * M_PI / 10.0), 
		      cos (2.0 * M_PI - l * M_PI / 10.0), 0.05);
	glEnd ();
	break;
      case 6:
	glNormal3f (0.0, 0.0, 1.0);
	glBegin (GL_POLYGON);
	glVertex3f (0.0, 1.0, 1.00);
	for (unsigned short l = 0; l < 20; ++l)
	  glVertex3f (sin (2.0 * M_PI - l * M_PI / 10.0), 
		      cos (2.0 * M_PI - l * M_PI / 10.0), 1.00);
	glEnd ();

	glBegin (GL_QUAD_STRIP);
	for (unsigned short l = 0; l < 21; ++l) {
	  glVertex3f (sin (2.0 * M_PI - l * M_PI / 10.0), 
		      cos (2.0 * M_PI - l * M_PI / 10.0), 0.05);
	  glVertex3f (sin (2.0 * M_PI - l * M_PI / 10.0), 
		      cos (2.0 * M_PI - l * M_PI / 10.0), 1.00);
	}
	glEnd ();

	glNormal3f (0.0, 0.0, -1.0);
	glBegin (GL_POLYGON);
	glVertex3f (0.0, 1.0, 0.05);
	for (unsigned short l = 0; l < 20; ++l)
	  glVertex3f (sin (2.0 * M_PI - l * M_PI / 10.0), 
		      cos (2.0 * M_PI - l * M_PI / 10.0), 0.05);
	glEnd ();
	break;
      }
      glEndList ();
    }
  }
  return;
}

Box::Box (ObjectShape s, ObjectColor c)
  : Object ("Box_" + Object::shape[s] + "_" + Object::color[c]) { 
  col = c;
  sh = s;
}

void Box::init () {
  return;
}

void Box::render () const {
  glMaterialfv (GL_FRONT, GL_AMBIENT, Object::ambient + col * 4);
  glMaterialfv (GL_FRONT, GL_DIFFUSE, Object::diffuse + col * 4);
  glMaterialfv (GL_FRONT, GL_SPECULAR, Object::specular + col * 4);
  glMaterialfv (GL_FRONT, GL_SHININESS, &Object::shine);
  glCallList (dl + sh);
  return;
}

Block::Block (ObjectShape s, ObjectColor c)
  : Object ("Block_" + Object::shape[s] + "_" + Object::color[s]) { 
  col = c;
  sh = s;
}

void Block::init () {
  return;
}

void Block::render () const {
  glMaterialfv (GL_FRONT, GL_AMBIENT, Object::ambient + col * 4);
  glMaterialfv (GL_FRONT, GL_DIFFUSE, Object::diffuse + col * 4);
  glMaterialfv (GL_FRONT, GL_SPECULAR, Object::specular + col * 4);
  glMaterialfv (GL_FRONT, GL_SHININESS, &Object::shine);
  glCallList (dl + sh);
  return;
}

Trap::Trap ()
  : Object ("Trap") { 
  col = BROWN;
  sh = DISC;
}

void Trap::init () {
  return;
}

void Trap::render () const {
  glMaterialfv (GL_FRONT, GL_AMBIENT, Object::ambient + col * 4);
  glMaterialfv (GL_FRONT, GL_DIFFUSE, Object::diffuse + col * 4);
  glMaterialfv (GL_FRONT, GL_SPECULAR, Object::specular + col * 4);
  glMaterialfv (GL_FRONT, GL_SHININESS, &Object::shine);
  glCallList (dl + sh);
  return;
}

Collect::Collect (ObjectColor c)
  : Object ("Collect_" + Object::color[c]) { 
  col = c;
  sh = DISC;
}

void Collect::init () {
  return;
}

void Collect::render () const {
  glMaterialfv (GL_FRONT, GL_AMBIENT, Object::ambient + col * 4);
  glMaterialfv (GL_FRONT, GL_DIFFUSE, Object::diffuse + col * 4);
  glMaterialfv (GL_FRONT, GL_SPECULAR, Object::specular + col * 4);
  glMaterialfv (GL_FRONT, GL_SHININESS, &Object::shine);
  glCallList (dl + sh);
  return;
}

Checkers::Checkers (ObjectColor c)
  : Object ("Checkers_" + Object::color[c]) { 
  col = c;
}

void Checkers::init () {
  return;
}

void Checkers::render () const {
  glMaterialfv (GL_FRONT, GL_AMBIENT, Object::ambient + col * 4);
  glMaterialfv (GL_FRONT, GL_DIFFUSE, Object::diffuse + col * 4);
  glMaterialfv (GL_FRONT, GL_SPECULAR, Object::specular + col * 4);
  glMaterialfv (GL_FRONT, GL_SHININESS, &Object::shine);
  glCallList (dl + CYL);
  return;
}

/// Local Variables:        ***
/// compile-command: "make" ***
/// End:                    ***
